/*
 * $Id$
 *
 * Serial Wire Debug Open Framework (BSD license).
 * Library Body File.
 *
 * Copyright (C) 2010, Tomasz Boleslaw CEDRO (http://www.tomek.cedro.info)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 3. Neither the name of the Tomasz Boleslaw CEDRO nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.*
 *
 * Written by Tomasz Boleslaw CEDRO <tomek.cedro@gmail.com>, 2010;
 *
 */

#include <libswd.h>
#include <string.h>

/* DATA GENERATED BELOW ARE TO BE SHIFTED OUT MSB FIRST, UNLIKE IN JTAG!
 * These functions generate a bistream packets for SWD communications.
 * Bistream is returned in byte array pointed by *buffer parameter.
 * Actual positive byte count is returned by the function.
 * Negative return value means an error. Zero means no data was generated.
 * If you need to swap bit direction, use bitswap functions provided below.
 * All functions name start with "swd_bitgen_" prefix.
 */

/* It will put both JTAG-DA and SW-DP into reset state. */
int swd_bitgen_reset(char *buffer){
    char rststr[8] = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};
    return sizeof(memcpy(buffer, rststr, sizeof(rststr)));
}

/* This will activate SW-DP. Remember about reset stream before and after. */
int swd_bitgen_switch_jtag_to_swd(char *buffer){
    char jtag2swd_seq[2] = {0x79, 0xE7};
    return sizeof(memcpy(buffer, jtag2swd_seq, sizeof(jtag2swd_seq)));
}

/* This will activate JTAG-DP. Remember about reset stream before and after. */
int swd_bitgen_switch_swd_to_jtag(char *buffer){
    char swd2jtag_seq[2] = {0x3C, 0xE7};
    return sizeof(memcpy(buffer, swd2jtag_seq, sizeof(swd2jtag_seq)));
}


/* BitSwap helper function that reverse bit order in *buffer. 
 * Least Significant Bit becomes Most Significant Bit.
 * Rersult contains only "count" bits for use with char or int easily.
 * This function operates on unsigned char (8-bit) data type.
 */
int swd_bitswap8(unsigned char *buffer, int bitcount){
 if (bitcount>8) return SWD_ERROR_BADPARAM;
 unsigned char bit, res=0; //res must be unsigned for proper shifting result
 #ifdef __SWDDEBUG__
 printf("|SWD_DEBUG: swd_bitswap8(%02X, %d);\n", *buffer, bitcount);
 #endif
 for (bit=0;bit<bitcount;bit++) {
  res=(res<<1)|(((*buffer>>bit)&1)?1:0); 
  #ifdef __SWDDEBUG__
  printf("|SWD_DEBUG: swd_bitswap8: in=%02X out=%02X bit=%d\n", *buffer, res, bit);
  #endif
 }
 *buffer=res;
 return SWD_OK;
}

/* BitSwap helper function that reverse bit order in *buffer. 
 * Least Significant Bit becomes Most Significant Bit.
 * Rersult contains only "count" bits for use with char or int easily.
 * This function operates on unsigned int (32-bit) data type.
 */
int swd_bitswap32(unsigned int *buffer, int bitcount){
 if (bitcount>32) return SWD_ERROR_BADPARAM;
 unsigned int bit, res=0; //res must be unsigned for proper shifting result
 #ifdef __SWDDEBUG__
 printf("|SWD_DEBUG: swd_bitswap32(%08X, %d);\n", *buffer, bitcount);
 #endif
 for (bit=0;bit<bitcount;bit++) {
  res=(res<<1)|(((*buffer>>bit)&1)?1:0); 
  #ifdef __SWDDEBUG__
  printf("|SWD_DEBUG: swd_bitswap32: in=%08X out=%08X bit=%d\n", *buffer, res, bit);
  #endif
 }
 *buffer=(int)res;
 return SWD_OK;
}


/* Command Queue helper functions.
 * All functions start with "swd_queue_" prefix. 
 */

/* Find first element of a queue pointed by *cmdq element. */
swd_cmd_t* swd_queue_find_root(swd_cmd_t *cmdq){
 swd_cmd_t *cmd=cmdq;
 while (cmd->prev!=NULL) cmd=cmd->prev;
 return cmd;
}

/* Find last element of a queue pointed by *cmdq element. */
swd_cmd_t* swd_queue_find_end(swd_cmd_t *cmdq){
 swd_cmd_t *cmd=cmdq;
 while (cmd->next!=NULL) cmd=cmd->next;
 return cmd;
}

/* Append element pointed by *cmd at the end of the quque pointed by *cmdq. */
int swd_queue_append(swd_cmd_t *cmdq, swd_cmd_t *cmd){
 if (cmdq==NULL || cmd==NULL) return SWD_ERROR_NULLPOINTER;
 if (cmdq->next != NULL){
  swd_cmd_t *lastcmd;
  lastcmd=swd_queue_find_end(cmdq);
  lastcmd->next=cmd;
  cmd->prev=lastcmd;
 } else {
  cmdq->next=cmd;
  cmd->prev=cmdq;
 }
 cmdq=cmd; 
 return 1;
}

/* Free memory / destroy queue pointed by the *cmdq element. */
void swd_queue_free(swd_cmd_t *cmdq){
 swd_cmd_t *cmd, *nextcmd;
 cmd=swd_cmd_queue_find_root(cmdq);
 while (cmd!=NULL) {
  nextcmd=cmd->next;
  free(cmd);
  cmd=nextcmd;
 }
}

/* Free memory / destroy / shorten queue head from root to the *cmdq element. */
void swd_queue_free_head(swd_cmd_t *cmdq){
 swd_cmd *cmdqroot, *nextcmd;
 cmdqroot=swd_cmd_queue_find_root(cmdq);
 while(cmdqroot!=cmdq){
  nextcmd=cmdqroot->next;
  free(cmdqroot);
  cmdqroot=nextcmd;
 }
}




/* SWD Commands Queue Elements Generator Routines.
 * These command quants are created in memory and can be easily appended
 * to the queue pointed by the *cmdq element.
 * All functions start with "swd_cmd_append_" prefix. */

/* Append SWD Request Packet Header to the context command queue.
 * Note that request is not validated here so bad request can be sent as well.*/
int swd_cmd_append_request(swd_ctx_t *swdctx, char request){
 if (swdctx==NULL) return SWD_ERROR_NULLCONEXT;
 swd_cmd_t *cmd;
 cmd=(swd_cmd_t *)calloc(1,sizeof(swd_cmd_t));
 if (cmd==NULL) return SWD_ERROR_OUTOFMEM;
 cmd->request=request;
 cmd->bits=SWD_REQUEST_BITLEN;
 cmd->type=SWD_CMDTYPE_MOSI;
 return swd_cmd_queue_append(swdctx->cmdq, cmd);
} 

/* Append SWD Turnaround to the context command queue. */ 
int swd_cmd_append_trn(swd_ctx_t *swdctx, int RnW){
 if (swdctx==NULL) return SWD_ERROR_NULLCONTEXT;
  swd_cmd_t *cmd;
  cmd=(swd_cmd_t *)calloc(1,sizeof(swd_cmd_t));
  if (cmd==NULL) return SWD_ERROR_OUTOFMEM;
  cmd->trn=RnW;
  cmd->bits=swdctx->params->trnlen;
  cmd->type=SWD_CMDTYPE_TRN;
  return swd_cmd_queue_append(swdctx->cmdq, cmd);
}

/* Append bus binary read bit-by-bit operation into context command queue.
 * Note that each command returns binary value into integer data element.
 * On error memory is freed and error code is returned.
 * Memory pointed by *data must be allocated prior calling this function. */
int swd_cmd_append_bitread(swd_ctx_t *swdctx, int *data, int count){
 if (swdctx==NULL || data==NULL || count<=0) return SWD_ERROR_BADPARAM;
 swd_cmd_t *cmd[];
 int i,rc,rq;
 for (i=0;i<=count;i++){
  cmd[i]=(swd_cmd_t *)calloc(1,sizeof(swd_cmd_t));
  rc=cmd[i];
  cmd[i]->misodata=&data[i];
  cmd[i]->bits=1;
  cmd[i]->type=SWD_CMDTYPE_BMISO;
  rq=swd_queue_append(swdctx->cmdq, cmd[i]);
  if (!rc || !rq){
   int j=i;
   for (j=i;j>=0;j--) free(cmd[j]);
   return SWD_ERROR_OUTOFMEM;
  }
 }
 return count;
}

/* Append 32-bit chunk + 1-bit parity read operation into context cmd queue. */
int swd_cmd_append_read(swd_ctx_t *swdctx, int *data){
 if (swdctx==NULL || data==NULL) return SWD_ERROR_NULLPOINTER;
 swd_cmd_t *cmd;
 cmd=(swd_cmd_t *)calloc(1,sizeof(swd_cmd_t));
 if (cmd==NULL) return SWD_ERROR_OUTOFMEM;
 cmd->misodata=&data;
 cmd->bits=33;
 cmd->type=SWD_CMDTYPE_MISO;
 return swd_queue_append(swdctx->cmdq, cmd);
}

/* Append n-chunk read operation into context command queue. */
int swd_cmd_append_nread(swd_ctx_t *swdctx, int *data, int count){
 if (swdctx==NULL || data==NULL || count<=0) return SWD_ERROR_BADPARAM;
 int i;
 for (i=0;i<=count;i++) if (swd_cmd_queue_append_read(cmdq, &data[i])<1) return SWD_ERROR_BADRESULT;
 return i;
}


int swd_cmd_append_write(swd_ctx_t *swdctx, int *data){
 if (swdctx==NULL || data==NULL) return SWD_ERROR_NULLPOINTER;
 swd_cmd_t *cmd;
 cmd=(swd_cmd_t *)calloc(1,sizeof(swd_cmd_t));
 if (cmd==NULL) return SWD_ERROR_OUTOFMEM;
 cmd->wdata=*data;
 cmd->bits=32;
 cmd->type=SWD_CMDTYPE_MOSI;
 return swd_cmd_queue_append(swdctx->cmdq, cmd);
}

int swd_cmd_queue_append_nwrite(swd_cmd *cmdq, int *data, int count){
 if (cmdq==NULL || data==NULL || count<1) return SWD_ERROR_BADPARAM;
  int i;
 for (i=0;i<count;i++) if (swd_cmd_queue_append_write(cmdq, &data[i])<1) return SWD_ERROR_BADRESULT;
 return i;
}

int swd_cmd_queue_append_ack(swd_cmd *cmdq){
 if (cmdq==NULL) return SWD_ERROR_NULLQUEUE;
 swd_cmd *cmd;
 cmd=(swd_cmd *)calloc(1,sizeof(swd_cmd)); 
 if (cmd==NULL) return SWD_ERROR_OUTOFMEM;  
 cmd->bits=3;
 cmd->type=SWD_CMDTYPE_ACK;
 return swd_cmd_queue_append(cmdq, cmd);
}

int swd_cmd_queue_append_trn_hosttx(swd_cmd *cmdq){
 return swd_cmd_queue_append_trn(cmdq, swd_trn_hosttx);
}

int swd_cmd_queue_append_trn_hostrx(swd_cmd *cmdq){
 return swd_cmd_queue_append_trn(cmdq, swd_trn_hostrx);
}

int swd_cmd_dpreset(swd_cmd *cmdq){
 char rststr[8] = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};
 int i;
 swd_cmd_queue_append_trn_hosttx(cmdq);
 for (i=0; i<sizeof(rststr);i++) swd_cmd_queue_append_request(cmdq, rststr[i]);
 return i;
}

int swd_cmd_jtag2swd(swd_cmd *cmdq){
 char jtag2swdseq[2] = {0x79, 0xE7};
 int i;
 swd_cmd_queue_append_trn_hosttx(cmdq);
 for (i=0; i<sizeof(jtag2swdseq);i++) swd_cmd_queue_append_request(cmdq, jtag2swdseq[i]);
 return i;
}



#ifdef __SWDHAVEMAIN__

int main(int argc, char argv[]){
 printf("Hello to SWD world, a simple SWD API test program by Tomek Cedro (http://www.tomek.cedro.info)\n");

 int a=0x01a0cede;
 unsigned char b=0x6F;
 printf("\nBitswap test:\n");
 printf("a size: %d bits, b size: %d bits\n", sizeof(a)*SWD_DATA_BYTESIZE, sizeof(b)*SWD_DATA_BYTESIZE);
 printf("Before bitswap: a=%08X , b=%02X \n", a, b);
 swd_bitswap32(&a, sizeof(a)*SWD_DATA_BYTESIZE);
 swd_bitswap8(&b, SWD_DATA_BYTESIZE);
 printf("After bitswap : a=%08X , b=%02X \n\n", a, b);
 return 0;
}
#endif
